﻿using System;
using System.Linq;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  [Item("SymbolFrequencyReinitializationStrategyController", "")]
  [StorableType("8F0B7020-A02F-491D-881A-5CE550426E5B")]
  public class SymbolFrequencyReinitializationStrategyController : InstrumentedOperator, IReinitializationStrategyController {

    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";
    private const string SymbolicExpressionTreeGrammarParameterName = "ClonedSymbolicExpressionTreeGrammar";
    private const string MinimumFrequencyParameterName = "MinimumFrequency";
    private const string LearningRateParameterName = "LearningRate";

    #region Parameter Properties
    public IScopeTreeLookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (IScopeTreeLookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }

    public IValueLookupParameter<ISymbolicExpressionGrammar> SymbolicExpressionTreeGrammarParameter {
      get { return (IValueLookupParameter<ISymbolicExpressionGrammar>)Parameters[SymbolicExpressionTreeGrammarParameterName]; }
    }

    public IFixedValueParameter<DoubleValue> MinimumFrequencyParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters[MinimumFrequencyParameterName]; }
    }

    public IFixedValueParameter<DoubleValue> LearningRateParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters[LearningRateParameterName]; }
    }
    #endregion

    #region Properties
    public double MinimumFrequency {
      get { return MinimumFrequencyParameter.Value.Value; }
      set { MinimumFrequencyParameter.Value.Value = value; }
    }

    public double LearningRate {
      get { return LearningRateParameter.Value.Value; }
      set { LearningRateParameter.Value.Value = value; }
    }
    #endregion


    #region Constructors
    public SymbolFrequencyReinitializationStrategyController() {
      Parameters.Add(new ScopeTreeLookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression tree whose length should be calculated."));
      Parameters.Add(new ValueLookupParameter<ISymbolicExpressionGrammar>(SymbolicExpressionTreeGrammarParameterName, "The tree grammar that defines the correct syntax of symbolic expression trees that should be created."));
      Parameters.Add(new FixedValueParameter<DoubleValue>(MinimumFrequencyParameterName, "Minimum Frequency for the controller to set to symbols.", new DoubleValue(0)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(LearningRateParameterName, "Learning Rate for how fast the frequency is adapted. Zero learning rate means no adaption, one means new frequency = symbol frequency.", new DoubleValue(0.1)));
    }

    private SymbolFrequencyReinitializationStrategyController(SymbolFrequencyReinitializationStrategyController original, Cloner cloner)
      : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolFrequencyReinitializationStrategyController(this, cloner);
    }

    [StorableConstructor]
    private SymbolFrequencyReinitializationStrategyController(StorableConstructorFlag _) : base(_) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey(MinimumFrequencyParameterName))
        Parameters.Add(new FixedValueParameter<DoubleValue>(MinimumFrequencyParameterName, "Minimum Frequency for the controller to set to symbols.", new DoubleValue(0)));
      if (!Parameters.ContainsKey(LearningRateParameterName))
        Parameters.Add(new FixedValueParameter<DoubleValue>(LearningRateParameterName, "Learning Rate for how fast the frequency is adapted. Zero learning rate means no adaption, one means new frequency = symbol frequency.", new DoubleValue(0.1)));
    }
    #endregion


    public override IOperation InstrumentedApply() {
      var trees = SymbolicExpressionTreeParameter.ActualValue;
      var grammar = SymbolicExpressionTreeGrammarParameter.ActualValue;
      double minimumFrequency = MinimumFrequency;
      double learningRate = LearningRate;

      var symbolFrequencies = SymbolicExpressionSymbolFrequencyAnalyzer.CalculateSymbolFrequencies(trees).ToDictionary(x => x.Key, x => x.Value);

      foreach (var s in grammar.AllowedSymbols) {
        if (s is ProgramRootSymbol || s is StartSymbol) {
          continue;
        }

        if (symbolFrequencies.TryGetValue(s.Name, out double f)) {
          double oldFrequency = s.InitialFrequency;
          double symbolFrequency = f;
          double newFrequency = oldFrequency + (symbolFrequency - oldFrequency) * learningRate;
          s.InitialFrequency = Math.Max(minimumFrequency, newFrequency);
        }
      }

      return base.InstrumentedApply();
    }
  }
}
